class InvalidFeatureBlockException(Exception):
    def __init__(self, spec_metadata, feature_name, block_ast_node, message, *args) -> None:
        super().__init__(*args)
        self.spec_metadata = spec_metadata
        self.feature_name = feature_name
        self.block_ast_node = block_ast_node
        self.message = message



import sys

from nimoy.runner.unittest_execution_framework import UnitTestExecutionFramework
from nimoy.spec_runner import SpecRunner


def main():
    execution_framework = UnitTestExecutionFramework()
    result = SpecRunner().run(execution_framework)
    sys.exit(not result.wasSuccessful())


if __name__ == '__main__':
    main()



import unittest


class UnitTestExecutionFramework:
    def __init__(self, stream=None):
        self.suite = unittest.TestSuite()
        self.stream = stream

    def append_test(self, test):
        self.suite.addTest(test)

    def run(self):
        return unittest.TextTestRunner(stream=self.stream, verbosity=2).run(self.suite)



from enum import Enum


class Types(Enum):
    EQUAL = 1
    NOT_EQUAL = 2
    LESS_THAN = 3
    LESS_THAN_EQUAL = 4
    GREATER_THAN = 5
    GREATER_THAN_EQUAL = 6
    IS = 7
    IS_NOT = 8
    IN = 9
    NOT_IN = 10
    MATCHES_REGEXP = 11



class MockAssertions:
    def assert_mock(self, number_of_invocations, mock, method, *args):
        if not hasattr(mock, method) and number_of_invocations > 0:
            raise AssertionError(method + " was never invoked") from None

        mocked_method = getattr(mock, method)

        if (number_of_invocations == 0) and (mocked_method.call_count == 0):
            return

        if (number_of_invocations >= 0) and (mocked_method.call_count != number_of_invocations):
            raise AssertionError(
                method + " was to be invoked " + str(number_of_invocations) + " times but was invoked " + str(
                    mocked_method.call_count)) from None

        for value, expected_value in zip(mocked_method.call_args[0], args):
            if expected_value != '__nimoy_argument_wildcard' and expected_value != value:
                raise AssertionError(
                    method + " expected argument " + str(expected_value) + " but was invoked with " + str(
                        value)) from None



class SpecReader:
    def __init__(self, resource_reader) -> None:
        super().__init__()
        self.resource_reader = resource_reader

    def read(self, spec_locations):
        def spec_contents():
            for spec_file_location in spec_locations:
                text = self.resource_reader.read(spec_file_location.spec_path)
                yield (spec_file_location, text)

        return spec_contents()



import argparse
import os

from nimoy.ast_tools import ast_chain
from nimoy.runner import fs_resource_reader
from nimoy.runner.spec_executor import SpecExecutor
from nimoy.runner.spec_finder import SpecFinder
from nimoy.runner.spec_loader import SpecLoader
from nimoy.runner.spec_reader import SpecReader


class SpecRunner:
    def run(self, execution_framework):
        spec_locations = SpecRunner._find_specs()
        spec_locations_and_contents = SpecRunner._read_specs(spec_locations)
        return SpecRunner._run_on_contents(execution_framework, spec_locations_and_contents)

    @staticmethod
    def _run_on_contents(execution_framework, spec_locations_and_contents):
        specs = SpecRunner._load_specs(spec_locations_and_contents)
        return SpecRunner._execute_specs(execution_framework, specs)

    @staticmethod
    def _find_specs():
        parser = argparse.ArgumentParser(prog='nimoy', description='Run a suite of Nimoy specs.')
        parser.add_argument('specs', metavar='S', type=str, nargs='*',
                            help="""A path to a spec file to execute or a directory to scan for spec files.
                            When naming a file it is possible to select which spec or feature to run. some_spec.py[::SpecName[::feature_name]]
                            """)
        args = parser.parse_args()
        suggested_locations = args.specs
        return SpecFinder(os.getcwd()).find(suggested_locations)

    @staticmethod
    def _read_specs(spec_locations):
        return SpecReader(fs_resource_reader).read(spec_locations)

    @staticmethod
    def _load_specs(spec_locations_and_contents):
        return SpecLoader(ast_chain).load(spec_locations_and_contents)

    @staticmethod
    def _execute_specs(execution_framework, specs):
        return SpecExecutor(execution_framework).execute(specs)



import difflib
from hamcrest import assert_that
from hamcrest import equal_to
from hamcrest import is_not
from hamcrest import less_than
from hamcrest import less_than_or_equal_to
from hamcrest import greater_than
from hamcrest import greater_than_or_equal_to
from hamcrest import is_
from hamcrest import is_in
from hamcrest import matches_regexp
from nimoy.compare.types import Types


class Compare:
    def compare(self, left, right, comparison_type_name):
        comparison_type = Types[comparison_type_name]
        try:
            self._select_comparison_method(left, right, comparison_type)
        except AssertionError as assertion_error:
            error_message = Compare.__reformat_error_message(assertion_error, left, right)
            raise AssertionError(error_message) from None

    def _select_comparison_method(self, left, right, comparison_type):
        if comparison_type == Types.EQUAL:
            assert_that(left, equal_to(right))
        elif comparison_type == Types.NOT_EQUAL:
            assert_that(left, is_not(equal_to(right)))
        elif comparison_type == Types.LESS_THAN:
            assert_that(left, less_than(right))
        elif comparison_type == Types.LESS_THAN_EQUAL:
            assert_that(left, less_than_or_equal_to(right))
        elif comparison_type == Types.GREATER_THAN:
            assert_that(left, greater_than(right))
        elif comparison_type == Types.GREATER_THAN_EQUAL:
            assert_that(left, greater_than_or_equal_to(right))
        elif comparison_type == Types.IS:
            assert_that(left, is_(right))
        elif comparison_type == Types.IS_NOT:
            assert_that(left, is_not(right))
        elif comparison_type == Types.IN:
            assert_that(left, is_in(right))
        elif comparison_type == Types.NOT_IN:
            assert_that(left, is_not(is_in(right)))
        elif comparison_type == Types.MATCHES_REGEXP:
            assert_that(left, matches_regexp(right))
        else:
            raise Exception()

    @staticmethod
    def __reformat_error_message(assertion_error, left, right):
        error_message = str(assertion_error)
        if Compare.__is_string(left) and Compare.__is_string(right):
            differ = difflib.Differ()
            diff = differ.compare(left.split('\n'), right.split('\n'))
            error_message = "%sHint:\n%s" % (error_message, '\n'.join(diff))
        return error_message

    @staticmethod
    def __is_string(value):
        return isinstance(value, str)

    @staticmethod
    def __is_list(value):
        return isinstance(value, list)



import _ast
import ast

from nimoy.ast_tools import ast_proxy
from nimoy.ast_tools.feature_blocks import FeatureBlockRuleEnforcer
from nimoy.ast_tools.feature_blocks import FeatureBlockTransformer


class FeatureRegistrationTransformer(ast.NodeTransformer):
    def __init__(self, spec_location, spec_metadata) -> None:
        super().__init__()
        self.spec_location = spec_location
        self.spec_metadata = spec_metadata

    def visit_FunctionDef(self, feature_node):
        if FeatureRegistrationTransformer._skip_feature(feature_node):
            return feature_node

        feature_name = feature_node.name
        if not feature_name.startswith('_'):

            feature_name_specified = hasattr(self.spec_location, 'feature_name')

            if not feature_name_specified or (
                    feature_name_specified and self.spec_location.feature_name == feature_name):
                self.spec_metadata.add_feature(feature_name)
                FeatureBlockTransformer(self.spec_metadata, feature_name).visit(feature_node)
                FeatureBlockRuleEnforcer(self.spec_metadata, feature_name, feature_node).enforce_tail_end_rules()

        feature_variables = self.spec_metadata.feature_variables.get(feature_name)
        if feature_variables:
            existing_arg_names = [existing_arg.arg for existing_arg in feature_node.args.args]

            for feature_variable in feature_variables:
                if feature_variable in existing_arg_names:
                    continue
                feature_node.args.args.append(_ast.arg(arg=feature_variable))
                feature_node.args.defaults.append(ast_proxy.ast_name_constant(value=None))

        if self._feature_has_a_where_function(feature_name):
            self._remove_where_function_from_node(feature_name, feature_node)
        return feature_node

    @staticmethod
    def _remove_where_function_from_node(feature_name, feature_node):
        where_function = FeatureRegistrationTransformer._locate_where_function_within_feature(feature_name,
                                                                                              feature_node)
        feature_node.body.remove(where_function)

    def _feature_has_a_where_function(self, feature_name):
        return self.spec_metadata.where_functions.get(feature_name)

    @staticmethod
    def _locate_where_function_within_feature(feature_name, feature_node):
        def _is_a_where_function(body_element):
            return hasattr(body_element, 'name') and body_element.name == feature_name + '_where'

        return next(body_element for body_element in feature_node.body if _is_a_where_function(body_element))

    @staticmethod
    def _skip_feature(feature_node):
        decorators = feature_node.decorator_list
        return any((hasattr(decorator, 'attr') and decorator.attr == 'skip') for decorator in decorators)



import ast

from nimoy.ast_tools.ast_metadata import SpecMetadata
from nimoy.ast_tools.features import FeatureRegistrationTransformer


class SpecTransformer(ast.NodeTransformer):
    def __init__(self, spec_location, spec_metadata) -> None:
        super().__init__()
        self.spec_location = spec_location
        self.spec_metadata = spec_metadata

    def visit_ClassDef(self, class_node):

        class_is_spec = class_node.name.endswith('Spec')

        if class_is_spec:
            has_spec_name = hasattr(self.spec_location, 'spec_name')
            if not has_spec_name or (has_spec_name and self.spec_location.spec_name == class_node.name):

                metadata = SpecMetadata(class_node.name)
                self._register_spec(metadata)
                FeatureRegistrationTransformer(self.spec_location, metadata).visit(class_node)

                for feature_name in metadata.where_functions:
                    feature = next(
                        feature for feature in class_node.body if
                        hasattr(feature, 'name') and feature_name == feature.name)
                    index_of_feature = class_node.body.index(feature)

                    index_to_insert_where = index_of_feature + 1
                    where_function_to_insert = metadata.where_functions[feature_name]

                    if (index_to_insert_where + 1) > len(class_node.body):
                        class_node.body.append(where_function_to_insert)
                    else:
                        class_node.body.insert(index_to_insert_where, where_function_to_insert)

        return class_node

    def _register_spec(self, metadata):
        self.spec_metadata.append(metadata)



import fnmatch
import os


class SpecFinder:
    def __init__(self, working_directory) -> None:
        super().__init__()
        self.working_directory = working_directory
        self.spec_locations = []

    def find(self, suggested_locations):
        if not suggested_locations:
            suggested_locations.append(self.working_directory)
        self._find_specs_in_suggested_locations(suggested_locations)

        return self.spec_locations

    def _find_specs_in_suggested_locations(self, suggested_locations):

        for suggested_location in suggested_locations:
            normalized_suggested_location = self._normalize_suggested_location(suggested_location)

            if 'spec.py' in normalized_suggested_location:
                self.spec_locations.append(Location(normalized_suggested_location))
            else:
                self._find_specs_in_directory(normalized_suggested_location)

    def _normalize_suggested_location(self, suggested_location):

        if os.path.isabs(suggested_location):
            return suggested_location
        return os.path.join(self.working_directory, suggested_location)

    def _find_specs_in_directory(self, directory):
        for root, dir_names, file_names in os.walk(directory):
            for filename in fnmatch.filter(file_names, '*spec.py'):
                self.spec_locations.append(Location(os.path.join(root, filename)))


class Location:
    def __init__(self, suggested_location):

        # Format may be:
        # - some_spec.py
        # - some_spec.py::SpecName
        # - some_spec.py::feature_name
        # - some_spec.py::SpecName::feature_name
        split_suggested_location = suggested_location.split("::")

        # some_spec.py
        if len(split_suggested_location) == 1:
            self.spec_path = split_suggested_location[0]

        # some_spec.py::SpecName or some_spec.py::feature_name
        if len(split_suggested_location) == 2:
            self.spec_path = split_suggested_location[0]

            # some_spec.py::SpecName
            if split_suggested_location[1][0].isupper():
                self.spec_name = split_suggested_location[1]

            # some_spec.py::feature_name
            else:
                self.feature_name = split_suggested_location[1]

        # some_spec.py::SpecName::feature_name
        if len(split_suggested_location) == 3:
            self.spec_path = split_suggested_location[0]
            self.spec_name = split_suggested_location[1]
            self.feature_name = split_suggested_location[2]



from unittest import TestCase
import copy
from nimoy.context.feature_block_context import FeatureBlock
from nimoy.compare.internal import Compare
from nimoy.assertions.mocks import MockAssertions
from nimoy.assertions.exceptions import ExceptionAssertions


class DataDrivenSpecification(type):
    @staticmethod
    def data_driver(test_method):
        def helper(*args, **kwargs):
            if not hasattr(args[0], test_method.__name__ + '_where'):
                return test_method(*args, **kwargs)

            where_method = getattr(args[0], test_method.__name__ + '_where')
            data_to_inject = DataDrivenSpecification._get_data_to_inject(where_method)
            for data_set in data_to_inject:
                copy_of_kwargs = copy.copy(kwargs)
                copy_of_kwargs.update(data_set)
                test_method(*args, **copy_of_kwargs)

        helper.__name__ = test_method.__name__

        return helper

    @staticmethod
    def _get_data_to_inject(where_function):
        data_to_inject = {}
        where_function(data_to_inject)

        iterable_data_to_inject = []
        for key in data_to_inject:
            values = data_to_inject[key]
            for index, value in enumerate(values):
                if (index + 1) > len(iterable_data_to_inject):
                    iterable_data_to_inject.append({})
                iterable_data_to_inject[index][key] = value
        return iterable_data_to_inject

    def __new__(mcs, clsname, superclasses, attributedict):
        for attribute in attributedict:
            DataDrivenSpecification.wrap_data_driven_method(attributedict, attribute)

        return type.__new__(mcs, clsname, superclasses, attributedict)

    @staticmethod
    def wrap_data_driven_method(attributedict, attribute):
        if callable(attribute) or attribute.startswith("_"):
            return

        attributedict[attribute] = DataDrivenSpecification.data_driver(attributedict[attribute])


class Specification(TestCase, metaclass=DataDrivenSpecification):

    def __init__(self, methodName='runTest'):
        super().__init__(methodName)
        self.thrown_exceptions = []

    def _feature_block_context(self, block_name):
        return FeatureBlock(block_name, self.thrown_exceptions)

    def _compare(self, left, right, comparison_type_name):
        Compare().compare(left, right, comparison_type_name)

    def _assert_mock(self, number_of_invocations, mock, method, *args):
        MockAssertions().assert_mock(number_of_invocations, mock, method, *args)

    def _exception_thrown(self, expected_exception_type):
        return ExceptionAssertions().assert_exception(self.thrown_exceptions, expected_exception_type)



from nimoy.ast_tools.feature_blocks import WHEN, THEN


class FeatureBlock:
    def __init__(self, block_type, thrown_exceptions) -> None:
        super().__init__()
        self.block_type = block_type
        self.thrown_exceptions = thrown_exceptions

    def __enter__(self):
        pass

    def __exit__(self, the_type, value, traceback):
        if self.block_type == WHEN and the_type:
            self.thrown_exceptions.append((the_type, value, traceback))
            return True
        if self.block_type == THEN and not the_type and self.thrown_exceptions:
            last_thrown_exception = self.thrown_exceptions.pop()
            if last_thrown_exception:
                raise last_thrown_exception[0](last_thrown_exception[1]).with_traceback(last_thrown_exception[2])



class SpecExecutor:
    def __init__(self, execution_framework) -> None:
        super().__init__()
        self.execution_framework = execution_framework

    def execute(self, specs):
        for spec_metadata in specs:
            spec_class_module = spec_metadata.owning_module

            for spec_feature in spec_metadata.features:
                self.execution_framework.append_test(spec_class_module(spec_feature))

        return self.execution_framework.run()



import _ast


class AST:

    @staticmethod
    def str(s):
        return _ast.Str(s)

    @staticmethod
    def name_constant(value=None):
        return _ast.NameConstant(value=value)

    @staticmethod
    def args(args=None):
        return _ast.arguments(
            args=args,
            kwonlyargs=[],
            kw_defaults=[],
            defaults=[])

    @staticmethod
    def num(n=None):
        return _ast.Num(n=n)



import _ast
import ast

from nimoy.ast_tools import ast_proxy
from nimoy.compare.types import Types


class ComparisonExpressionTransformer(ast.NodeTransformer):
    def __init__(self) -> None:
        super().__init__()
        self.comparator_methods = {
            _ast.Eq: Types.EQUAL,
            _ast.NotEq: Types.NOT_EQUAL,
            _ast.Lt: Types.LESS_THAN,
            _ast.LtE: Types.LESS_THAN_EQUAL,
            _ast.Gt: Types.GREATER_THAN,
            _ast.GtE: Types.GREATER_THAN_EQUAL,
            _ast.Is: Types.IS,
            _ast.IsNot: Types.IS_NOT,
            _ast.In: Types.IN,
            _ast.NotIn: Types.NOT_IN,
            _ast.MatMult: Types.MATCHES_REGEXP,
        }

    def visit_Expr(self, expression_node):
        value = expression_node.value
        if not isinstance(value, _ast.Compare):
            if isinstance(value, _ast.BinOp):
                if hasattr(value, 'op') and not isinstance(value.op, _ast.MatMult):
                    return expression_node
            else:
                return expression_node

        left_value = value.left

        # TODO: Support multiple comparators (1 < 2 < 3). This may be tricky because one expression of multiple
        if hasattr(value, 'op'):
            comparison_operation = value.op
            right_value = value.right
        else:
            comparison_operation = value.ops[0]
            right_value = value.comparators[0]
        comparison_operation_type = type(comparison_operation)
        internal_comparison_type = self.comparator_methods[comparison_operation_type]

        expression_node.value = _ast.Call(
            func=_ast.Attribute(
                value=_ast.Name(id='self', ctx=_ast.Load()),
                attr='_compare',
                ctx=_ast.Load()
            ),
            args=[left_value, right_value, ast_proxy.ast_str(s=internal_comparison_type.name)],
            keywords=[]
        )
        return expression_node


class ThrownExpressionTransformer(ast.NodeTransformer):

    def visit_Call(self, expression_node):
        if isinstance(expression_node, _ast.Call):
            if hasattr(expression_node.func, 'id') and expression_node.func.id == 'thrown':
                expected_exception = expression_node.args[0]
                expression_node = _ast.Call(
                    func=_ast.Attribute(
                        value=_ast.Name(id='self', ctx=_ast.Load()),
                        attr='_exception_thrown',
                        ctx=_ast.Load()
                    ),
                    args=[expected_exception],
                    keywords=[]
                )
        return expression_node


class MockBehaviorExpressionTransformer(ast.NodeTransformer):
    def visit_Expr(self, expression_node):
        value = expression_node.value
        if not isinstance(value, _ast.BinOp) or not (
                isinstance(value.op, _ast.RShift) or isinstance(value.op, _ast.LShift)):
            return expression_node

        mock_attribute = 'return_value'
        if isinstance(value.op, _ast.LShift):
            mock_attribute = 'side_effect'

        return _ast.Assign(
            targets=[
                _ast.Attribute(
                    attr=mock_attribute,
                    ctx=_ast.Store(),
                    value=value.left.func
                )
            ],
            value=value.right
        )


class MockAssertionTransformer(ast.NodeTransformer):
    def visit_Expr(self, expression_node):
        value = expression_node.value
        if not isinstance(value, _ast.BinOp) or not isinstance(value.op, _ast.Mult) or not isinstance(value.right,
                                                                                                      _ast.Call):
            return expression_node

        number_of_invocations = value.left
        if MockAssertionTransformer._value_is_a_wildcard(number_of_invocations):
            number_of_invocations = ast_proxy.ast_num(n=-1)
        target_mock = value.right.func.value
        target_method = ast_proxy.ast_str(s=value.right.func.attr)

        list_of_arguments = [MockAssertionTransformer._transform_arg_if_wildcard(x) for x in value.right.args]
        spread_list_of_arguments = _ast.Starred(value=_ast.List(elts=list_of_arguments, ctx=_ast.Load()),
                                                ctx=_ast.Load())
        expression_node.value = _ast.Call(
            func=_ast.Attribute(
                value=_ast.Name(id='self', ctx=_ast.Load()),
                attr='_assert_mock',
                ctx=_ast.Load()
            ),
            args=[number_of_invocations, target_mock, target_method, spread_list_of_arguments],
            keywords=[]
        )
        return expression_node

    @staticmethod
    def _value_is_a_wildcard(value):
        return isinstance(value, _ast.Name) and value.id == '_'

    @staticmethod
    def _transform_arg_if_wildcard(arg):
        if MockAssertionTransformer._value_is_a_wildcard(arg):
            return ast_proxy.ast_str(s='__nimoy_argument_wildcard')

        return arg



import _ast
import ast
import copy

from nimoy.ast_tools import ast_proxy
from nimoy.ast_tools.expression_transformer import ComparisonExpressionTransformer, MockAssertionTransformer, \
    ThrownExpressionTransformer, MockBehaviorExpressionTransformer
from nimoy.runner.exceptions import InvalidFeatureBlockException

SETUP = 'setup'
GIVEN = 'given'
WHEN = 'when'
THEN = 'then'
EXPECT = 'expect'
WHERE = 'where'
BLOCK_NAMES = [SETUP, GIVEN, WHEN, THEN, EXPECT, WHERE]


class WhereBlockFunctions:
    def __init__(self, spec_metadata, feature_name) -> None:
        super().__init__()
        self.spec_metadata = spec_metadata
        self.feature_name = feature_name

    def assign_variables_to_the_context(self, where_function_ast_node):
        block_body = where_function_ast_node.body
        first_expression = block_body[0]
        if isinstance(first_expression, _ast.Assign):
            self._assign_list_form_variables(where_function_ast_node)
        elif hasattr(first_expression, 'value') and isinstance(first_expression.value, _ast.BinOp):
            self._assign_matrix_form_variables(where_function_ast_node)

    def _assign_list_form_variables(self, where_function_ast_node):
        copy_of_body = copy.deepcopy(where_function_ast_node.body)
        where_function_ast_node.body = []
        for assignment_expression in copy_of_body:
            variable_name = assignment_expression.targets[0].id
            self.spec_metadata.add_feature_variable(self.feature_name, variable_name)
            variable_values = assignment_expression.value
            where_function_ast_node.body.append(
                _ast.Assign(
                    targets=[
                        _ast.Subscript(
                            value=_ast.Name(id='injectable_values', ctx=_ast.Load()),
                            slice=_ast.Index(value=ast_proxy.ast_str(s=variable_name)),
                            ctx=_ast.Store()
                        )
                    ],
                    value=variable_values
                ))

    def _assign_matrix_form_variables(self, where_function_ast_node):
        copy_of_body = copy.deepcopy(where_function_ast_node.body)
        where_function_ast_node.body = []

        variables_and_values = WhereBlockFunctions._get_variables_and_values(copy_of_body)

        # We might be screwing with line numbers here
        for variable_name, variable_values in variables_and_values.items():
            self.spec_metadata.add_feature_variable(self.feature_name, variable_name)
            where_function_ast_node.body.append(
                _ast.Assign(
                    targets=[
                        _ast.Subscript(
                            value=_ast.Name(id='injectable_values', ctx=_ast.Load()),
                            slice=_ast.Index(value=ast_proxy.ast_str(s=variable_name)),
                            ctx=_ast.Store()
                        )
                    ],
                    value=_ast.List(elts=variable_values, ctx=_ast.Load())
                ))

    @staticmethod
    def _get_variables_and_values(copy_of_body):
        variable_names_row = copy_of_body[0].value
        variable_names = []
        WhereBlockFunctions._collect_variable_names_recursively(variable_names_row, variable_names)
        variables_and_values = {variable_names[i]: [] for i in range(len(variable_names))}
        for values_row in copy_of_body[1:]:
            variable_values = []
            WhereBlockFunctions._collect_variable_values_recursively(values_row.value, variable_values)
            for index, value in enumerate(variable_values):
                variables_and_values[variable_names[index]].append(value)
        return variables_and_values

    @staticmethod
    def _collect_variable_names_recursively(binary_op_node, variable_names):
        if isinstance(binary_op_node.left, _ast.BinOp):
            WhereBlockFunctions._collect_variable_names_recursively(binary_op_node.left, variable_names)
        else:
            variable_names.append(WhereBlockFunctions._get_variable_name(binary_op_node.left))
        variable_names.append(WhereBlockFunctions._get_variable_name(binary_op_node.right))

    @staticmethod
    def _get_variable_name(variable_object):
        if hasattr(variable_object, 's'):
            return variable_object.s
        return variable_object.id

    @staticmethod
    def _collect_variable_values_recursively(binary_op_node, variable_values):
        if isinstance(binary_op_node.left, _ast.BinOp):
            WhereBlockFunctions._collect_variable_values_recursively(binary_op_node.left, variable_values)
        else:
            variable_values.append(binary_op_node.left)
        variable_values.append(binary_op_node.right)


class FeatureBlockRuleEnforcer:
    def __init__(self, spec_metadata, feature_name, block_ast_node) -> None:
        super().__init__()
        self.spec_metadata = spec_metadata
        self.feature_name = feature_name
        self.block_ast_node = block_ast_node

    def enforce_addition_rules(self, block_type):
        existing_blocks = self.spec_metadata.feature_blocks.get(self.feature_name)

        if block_type in [SETUP, GIVEN]:
            if existing_blocks:
                if any([(existing_block in [SETUP, GIVEN]) for existing_block in existing_blocks]):
                    raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                       'Each feature may only have a single setup/given block')

                raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                   'setup/given blocks may only appear at the start of the feature')

        if block_type == THEN:
            if (existing_blocks and existing_blocks[-1] != WHEN) or not existing_blocks:
                raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                   'when blocks may only succeed then blocks')

        if block_type != THEN:
            if existing_blocks:
                if existing_blocks[-1] == WHEN:
                    raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                       'when blocks may be succeeded only by then blocks')

        if existing_blocks:
            if existing_blocks[-1] == WHERE:
                raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                   'No blocks may succeeded where blocks')

    def enforce_tail_end_rules(self):
        existing_blocks = self.spec_metadata.feature_blocks.get(self.feature_name)
        if existing_blocks:
            if existing_blocks[-1] == WHEN:
                raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                   'when blocks must be succeeded by a then block')
            if existing_blocks[-1] in [SETUP, GIVEN]:
                raise InvalidFeatureBlockException(self.spec_metadata, self.feature_name, self.block_ast_node,
                                                   'Features cannot end with a setup or a given block')


class FeatureBlockTransformer(ast.NodeTransformer):
    def __init__(self, spec_metadata, feature_name) -> None:
        super().__init__()
        self.spec_metadata = spec_metadata
        self.feature_name = feature_name

    def visit_With(self, with_node):

        if FeatureBlockTransformer._is_feature_block(with_node):
            block_type = with_node.items[0].context_expr.id
            FeatureBlockRuleEnforcer(self.spec_metadata, self.feature_name, with_node).enforce_addition_rules(
                block_type)
            self.spec_metadata.add_feature_block(self.feature_name, block_type)

            if block_type != WHERE:
                FeatureBlockTransformer._replace_with_block_context(with_node, block_type)
                if block_type == WHEN:
                    MockBehaviorExpressionTransformer().visit(with_node)

                if block_type in [THEN, EXPECT]:
                    ComparisonExpressionTransformer().visit(with_node)

                if block_type == THEN:
                    MockAssertionTransformer().visit(with_node)
                    ThrownExpressionTransformer().visit(with_node)
                return with_node

            where_function = self._replace_where_block_with_function(with_node)
            WhereBlockFunctions(self.spec_metadata, self.feature_name).assign_variables_to_the_context(where_function)
            self.spec_metadata.add_where_function(self.feature_name, copy.deepcopy(where_function))
            return where_function

    @staticmethod
    def _is_feature_block(with_node):
        if len(with_node.items) != 1:
            return False

        items = with_node.items[0]
        if not items.context_expr:
            return False

        if not items.context_expr.id:
            return False

        return items.context_expr.id in BLOCK_NAMES

    @staticmethod
    def _replace_with_block_context(with_node, block_type):
        with_node.items[0].context_expr = _ast.Call(
            func=_ast.Attribute(value=_ast.Name(id='self', ctx=_ast.Load()), attr='_feature_block_context',
                                ctx=_ast.Load()), args=[ast_proxy.ast_str(s=block_type)], keywords=[])

    def _replace_where_block_with_function(self, with_node):
        return _ast.FunctionDef(name=self.feature_name + '_where',
                                args=ast_proxy.ast_args([_ast.arg(arg='self'), _ast.arg(arg='injectable_values')]),
                                body=copy.deepcopy(with_node.body),
                                decorator_list=[],
                                returns=None)



class ExceptionAssertions:
    def assert_exception(self, thrown_exceptions, expected_exception_type):
        if not thrown_exceptions:
            raise AssertionError("Expected an exception of type '%s' to be thrown" % expected_exception_type.__name__)

        thrown_exception = thrown_exceptions.pop()
        if not isinstance(thrown_exception[1], expected_exception_type):
            raise AssertionError("Expected an exception of type '%s' but found '%s'" % (
                expected_exception_type.__name__, thrown_exception[0].__name__))

        return thrown_exception



class SpecMetadata:
    def __init__(self, name) -> None:
        super().__init__()
        self.name = name
        self.owning_module = None
        self.features = []
        self.feature_blocks = {}
        self.feature_variables = {}
        self.where_functions = {}

    def set_owning_module(self, owning_module):
        self.owning_module = owning_module

    def add_feature(self, feature_name):
        self.features.append(feature_name)
        self.feature_blocks[feature_name] = []
        self.feature_variables[feature_name] = {}

    def add_feature_block(self, feature_name, block_type):
        self.feature_blocks[feature_name].append(block_type)

    def add_feature_variable(self, feature_name, variable_name):
        if not self.feature_variables[feature_name]:
            self.feature_variables[feature_name] = []
        self.feature_variables[feature_name].append(variable_name)

    def add_where_function(self, feature_name, where_function):
        self.where_functions[feature_name] = where_function



from setuptools import setup

classifiers = [
                  'Development Status :: 5 - Production/Stable',
                  'Intended Audience :: Developers',
                  'License :: OSI Approved :: Apache Software License',
                  'Operating System :: POSIX',
                  'Operating System :: MacOS :: MacOS X',
                  'Topic :: Software Development :: Testing',
                  'Topic :: Software Development :: Libraries',
                  'Topic :: Utilities',
              ] + [
                  ('Programming Language :: Python :: %s' % x) for x in '3 3.3 3.4 3.5 3.6 3.7 3.8 3.9'.split()
              ]

with open('README.rst') as read_me:
    long_description = read_me.read()

setup(
    name='nimoy-framework',
    version='1.0.2',
    description='A testing and specification framework for Python 3, heavily inspired by the Spock Framework',
    long_description=long_description,
    url='https://github.com/browncoat-ninjas/nimoy',
    license='Apache License',
    platforms=['unix', 'linux', 'osx'],
    author='Noam Tenne, Yoav Luft',
    author_email='noam@10ne.org',
    entry_points={'console_scripts': ['nimoy = nimoy.main:main']},
    classifiers=classifiers,
    keywords="test unittest specification",
    packages=['nimoy', 'nimoy.assertions', 'nimoy.ast_tools', 'nimoy.context', 'nimoy.runner', 'nimoy.compare'],
    install_requires=['pyhamcrest==2.0.2', 'urllib3==1.26.2'],
)



import ast


class SpecLoader:
    def __init__(self, ast_chain) -> None:
        super().__init__()
        self.ast_chain = ast_chain

    def load(self, spec_locations_and_contents):
        def specs():
            for spec_location, text in spec_locations_and_contents:
                node = ast.parse(text, mode='exec')

                metadata_of_specs_from_node = self.ast_chain.apply(spec_location, node)
                ast.fix_missing_locations(node)
                compiled = compile(node, spec_location.spec_path, 'exec')

                spec_namespace = {}
                exec(compiled, spec_namespace)

                for spec_metadata in metadata_of_specs_from_node:
                    spec_metadata.set_owning_module(spec_namespace[spec_metadata.name])
                    yield spec_metadata

        return specs()



import sys

IS_3_8 = sys.version_info >= (3, 8)
if IS_3_8:
    from .ast_3_8 import AST
else:
    from .ast_3_7 import AST


def ast_str(s):
    return AST.str(s)


def ast_name_constant(value=None):
    return AST.name_constant(value)


def ast_args(args=None):
    return AST.args(args)


def ast_num(n=None):
    return AST.num(n)
